#
# Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""A module providing downloaders of bank statements and possibly other resources."""

import logging
from abc import ABC, abstractmethod
from datetime import date, timedelta
from enum import Enum, unique
from pathlib import Path
from typing import Any, Iterable, List, Mapping, Optional, Tuple, Union
from urllib.parse import urljoin
from warnings import warn

import requests
import zeep

# ServiceProxy is in different modules in zeep2 and zeep3. We need to support both.
try:
    from zeep.proxy import ServiceProxy
except ImportError:
    from zeep.client import ServiceProxy


LOGGER = logging.getLogger(__name__)


class TellerDownloadError(Exception):
    """Error raised by Teller when statement download fails."""


class BankStatementDownloader(ABC):
    """Abstract class for statement downloaders."""

    def __init__(self, base_url: str, timeout: int = 3):
        if not base_url.startswith('https://'):
            warn('Sensitive data is transferred over possibly insecure channel. Consider HTTPS.', UserWarning)
        self.base_url = base_url
        self.timeout = timeout

    def get_statements(self, start_date: date, end_date: date) -> Iterable[str]:
        """Download the bank statements. DO NOT LOG EXCEPTIONS raised by this method. It may contain sensitive data."""
        self._check_arguments(start_date, end_date)
        return self._download_data(start_date, end_date)

    def _check_arguments(self, start_date: date, end_date: date) -> None:
        if not start_date <= end_date:
            raise ValueError('Start date is greater than end date.')

    @abstractmethod
    def _download_data(self, start_date: date, end_date: date) -> Iterable[str]:
        """Download content from a bank."""


class HttpStatementDownloader(BankStatementDownloader):
    """Download bank statements over http."""

    def _download_data(self, start_date: date, end_date: date) -> Iterable[str]:
        """Download the bank statements. DO NOT LOG EXCEPTIONS raised by this method. It may contain sensitive data."""
        try:
            url = self._compose_url(start_date, end_date)
            params = self._get_params(start_date, end_date)
            response = requests.get(url, params=params, timeout=self.timeout)
            response.raise_for_status()
        except Exception:
            raise TellerDownloadError('Statement could not be retrieved.')
        return [response.text]

    def _compose_url(self, start_date: date, end_date: date) -> str:
        """Compose the bank specific URL from where the statement should be retrieved."""
        return self.base_url

    def _get_params(self, start_date: date, end_date: date) -> Optional[Mapping[str, Any]]:
        """Collect the bank specific request parametes."""
        return None


class FioStatementDownloader(HttpStatementDownloader):
    """Downloader for Fio bank statements."""

    date_format = '%Y-%m-%d'
    path_format = 'ib_api/rest/periods/{password}/{start}/{end}/transactions.xml'

    def __init__(self, base_url: str, password: str, **kwargs):
        super().__init__(base_url, **kwargs)
        self.password = password

    def _compose_url(self, start_date: date, end_date: date) -> str:
        start = start_date.strftime(self.date_format)
        end = end_date.strftime(self.date_format)
        path = self.path_format.format(password=self.password, start=start, end=end)
        return urljoin(self.base_url, path)


class RaiffeisenStatementDownloader(HttpStatementDownloader):
    """Downloader for Raiffeisen bank statements."""

    date_format = '%d.%m.%Y'
    path_format = 'ibs/eshop/payments-list'

    def __init__(self, base_url: str, shopname: str, creditaccount: int, creditbank: int, password: str,
                 **kwargs):
        super().__init__(base_url, **kwargs)
        self.shopname = shopname
        self.creditaccount = creditaccount
        self.creditbank = creditbank
        self.password = password

    def _download_data(self, start_date: date, end_date: date) -> Iterable[str]:
        statements = super()._download_data(start_date, end_date)
        return [self._strip_empty_text(text) for text in statements]

    def _compose_url(self, start_date: date, end_date: date) -> str:
        return urljoin(self.base_url, self.path_format)

    def _get_params(self, start_date: date, end_date: date) -> Optional[Mapping[str, Any]]:
        params = {
            'shopname': self.shopname,
            'creditaccount': self.creditaccount,
            'creditbank': self.creditbank,
            'password': self.password,
            'paidfrom': start_date.strftime(self.date_format),
            'paidto': end_date.strftime(self.date_format),
            'listtype': 'PLAIN',
        }
        return params

    def _strip_empty_text(self, text: str) -> str:
        """Turn string consisting of whitespace only to empty string.

        When no payment suits the filter criteria the bank returns a file consisting of one space. We prefer empty files
        in this case.
        """
        if text.strip() == '':
            return ''
        else:
            return text


class CSOBStatementDownloader(BankStatementDownloader):
    """Downloader for CSOB bank statements."""

    @unique
    class FileStatus(str, Enum):
        NOT_READY = 'R'
        READY = 'D'
        ERROR = 'F'

    def __init__(self, base_url: str, wsdl_file: Union[str, Path], contract_number: str = None,
                 cert_path: Union[str, Path] = None, key_path: Union[str, Path] = None, **kwargs):
        super().__init__(base_url, **kwargs)
        self.wsdl_file = str(wsdl_file)
        self.contract_number = contract_number
        if cert_path is not None and key_path is not None:
            self.cert = (str(cert_path), str(key_path))  # type: Union[Tuple[str, str], None]
        else:
            self.cert = None

    def _download_data(self, start_date: date, end_date: date) -> Iterable[str]:
        statement_urls = self._get_statement_urls(start_date, end_date)
        for url in statement_urls:
            try:
                response = requests.get(url, cert=self.cert)
                response.raise_for_status()
                yield response.text
            except Exception:
                raise TellerDownloadError('Statement at {} could not be obtained.'.format(url))

    def _get_statement_urls(self, start_date: date, end_date: date) -> List[str]:
        service = self._create_service()

        # CreatedBefore is exclusive so we add one day
        filter_params = {'FileTypes': ['VYPIS'],
                         'CreatedAfter': start_date,
                         'CreatedBefore': end_date + timedelta(days=1)}
        try:
            response = service.GetDownloadFileList_v2(self.contract_number, Filter=filter_params)
        except Exception:
            raise TellerDownloadError('Could not obtain file list.')
        files_details = response.FileList.FileDetail if response.FileList else []
        urls = []
        for detail in files_details:
            if detail.Filename.lower().endswith('.xml'):
                if detail.Status == self.FileStatus.READY:
                    urls.append(detail.Url)
                elif detail.Status == self.FileStatus.NOT_READY:
                    LOGGER.warning('Could not get file {} - file not ready.'.format(detail.Filename))
                elif detail.Status == self.FileStatus.ERROR:
                    raise TellerDownloadError('Could not get file {} - error on the bank side.'.format(detail.Filename))
                else:
                    raise TellerDownloadError('Unknown status of file {}.'.format(detail.Filename))
        return urls

    def _create_service(self) -> ServiceProxy:
        session = requests.Session()
        session.cert = self.cert
        transport = zeep.transports.Transport(session=session)
        client = zeep.Client(self.wsdl_file, transport=transport)
        service = client.create_service('{http://ceb-bc.csob.cz/CEBBCWS}CEBBCWSBinding', self.base_url)
        return service
