#
# Copyright (C) 2020-2021  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.
from datetime import date, timedelta
from pathlib import Path
from unittest import TestCase
from unittest.mock import Mock, patch, sentinel
from urllib.parse import urlencode

import responses
import zeep

from teller.downloaders import (BankStatementDownloader, CSOBStatementDownloader, FioStatementDownloader,
                                HttpStatementDownloader, RaiffeisenStatementDownloader, TellerDownloadError)


class DummyStatementDownloader(BankStatementDownloader):
    def _download_data(self, start_date, end_date):
        pass  # pragma: no cover


class TestBankStatementDownloader(TestCase):

    def test_enforce_https(self):
        with self.assertWarnsRegex(UserWarning, 'Sensitive data is transferred'):
            DummyStatementDownloader('http://bank.test', 1)

    def test_check_arguments(self):
        downloader = DummyStatementDownloader('https://bank.test', 1)
        with self.assertRaisesRegex(ValueError, 'Start date is greater than end date'):
            downloader.get_statements(start_date=date(2020, 10, 1), end_date=date(1999, 1, 1))


class TestHttpStatementDownloader(TestCase):
    url = 'https://bank.test'
    password = 'letmein'
    start_date = date(2020, 9, 3)
    end_date = date(2020, 10, 2)

    def test_download_data(self):
        downloader = HttpStatementDownloader(self.url)

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, self.url, body='It works!', status=200, content_type='text/plain')
            result = downloader.get_statements(self.start_date, self.end_date)
        self.assertEqual(['It works!'], result)

    def test_statement_not_found(self):
        downloader = HttpStatementDownloader(self.url)

        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, self.url, body='Not found.', status=404, content_type='text/plain')
            with self.assertRaisesRegex(TellerDownloadError, 'Statement could not be retrieved'):
                downloader.get_statements(self.start_date, self.end_date)


class TestFioStatementDownloader(TestCase):
    url = 'https://bank.test'
    password = 'letmein'
    start_date = date(2020, 9, 3)
    end_date = date(2020, 10, 2)

    def test_compose_url(self):
        downloader = FioStatementDownloader(self.url, self.password)

        result = downloader._compose_url(self.start_date, self.end_date)

        full_url = '{}/ib_api/rest/periods/{}/2020-09-03/2020-10-02/transactions.xml'.format(self.url, self.password)
        self.assertEqual(result, full_url)

    def test_get_statements(self):
        downloader = FioStatementDownloader(self.url, self.password)
        statement = 'I am a fish.'

        full_url = '{}/ib_api/rest/periods/{}/2020-09-03/2020-10-02/transactions.xml'.format(self.url, self.password)
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, full_url, body=statement, status=200, content_type='text/plain')
            result = downloader.get_statements(self.start_date, self.end_date)

        self.assertEqual(result, [statement])


class TestRaiffeisenStatementDownloader(TestCase):
    url = 'https://bank.test'
    password = 'letmein'
    shop = 'Name'
    account = 123
    bank = 1001
    start_date = date(2020, 9, 3)
    end_date = date(2020, 10, 2)

    def test_compose_url(self):
        downloader = RaiffeisenStatementDownloader(self.url, self.shop, self.account, self.bank, self.password)

        result = downloader._compose_url(self.start_date, self.end_date)

        full_url = '{}/ibs/eshop/payments-list'.format(self.url)
        self.assertEqual(result, full_url)

    def test_get_statements(self):
        downloader = RaiffeisenStatementDownloader(self.url, self.shop, self.account, self.bank, self.password)
        statement = 'I am a fish.'

        params = {'shopname': self.shop,
                  'creditaccount': str(self.account),
                  'creditbank': str(self.bank),
                  'password': self.password,
                  'listtype': 'PLAIN',
                  'paidfrom': '03.09.2020',
                  'paidto': '02.10.2020'}
        full_url = '{}/ibs/eshop/payments-list?{}'.format(self.url, urlencode(params))
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, full_url, body=statement, status=200, content_type='text/plain')
            result = downloader.get_statements(self.start_date, self.end_date)

        self.assertEqual(result, [statement])

    def test_empty_statements(self):
        # When no payment suits the filter criteria the bank returns a file consisting of one space.
        downloader = RaiffeisenStatementDownloader(self.url, self.shop, self.account, self.bank, self.password)

        url = '{}/ibs/eshop/payments-list'.format(self.url)
        with responses.RequestsMock() as rsps:
            rsps.add(responses.GET, url, body=' ', status=200, content_type='text/plain', match_querystring=False)
            rsps.add(responses.GET, url, body='\n', status=200, content_type='text/plain', match_querystring=False)
            rsps.add(responses.GET, url, body=' \n', status=200, content_type='text/plain', match_querystring=False)
            rsps.add(responses.GET, url, body=' abc\n', status=200, content_type='text/plain', match_querystring=False)
            result_1 = downloader.get_statements(self.start_date, self.end_date)
            result_2 = downloader.get_statements(self.start_date, self.end_date)
            result_3 = downloader.get_statements(self.start_date, self.end_date)
            result_4 = downloader.get_statements(self.start_date, self.end_date)

        self.assertEqual(result_1, [''])
        self.assertEqual(result_2, [''])
        self.assertEqual(result_3, [''])
        self.assertEqual(result_4, [' abc\n'])


class TestCSOBStatementDownloader(TestCase):
    url = 'https://bank.test/cebbc/api'
    wsdl = 'teller/tests/data/cebbc-wsdl/CEBBCWS.wsdl'
    contract = '12345'
    start_date = date(2020, 9, 3)
    end_date = date(2020, 10, 2)
    cert = Path('path/to/cert')
    key = Path('path/to/key')

    FileDetail = zeep.Client(wsdl).type_factory('http://ceb-bc.csob.cz/CEBBCWS/GetDownloadFileList_v2').FileDetail_v1

    def test_get_statements(self):
        downloader = CSOBStatementDownloader(self.url, self.wsdl)
        with patch.object(downloader, '_get_statement_urls', autospec=True) as mock_method:
            mock_method.return_value = [self.url + '/file_1', self.url + '/file_2']
            with responses.RequestsMock() as rsps:
                rsps.add(responses.GET, self.url + '/file_1', body='<content_1>', status=200, content_type='text/plain')
                rsps.add(responses.GET, self.url + '/file_2', body='<content_2>', status=200, content_type='text/plain')
                result = list(downloader.get_statements(self.start_date, self.end_date))
        self.assertEqual(['<content_1>', '<content_2>'], result)

    def test_get_statements_missing(self):
        downloader = CSOBStatementDownloader(self.url, self.wsdl)
        with patch.object(downloader, '_get_statement_urls', autospec=True) as mock_method:
            mock_method.return_value = [self.url + '/file_1']
            with responses.RequestsMock() as rsps:
                rsps.add(responses.GET, self.url + '/file_1', body='Nothing', status=404, content_type='text/plain')
                expected_message = 'Statement at https://bank.test/cebbc/api/file_1 could not be obtained'
                with self.assertRaisesRegex(TellerDownloadError, expected_message):
                    list(downloader.get_statements(self.start_date, self.end_date))

    def test_download_data_certs(self):
        downloader = CSOBStatementDownloader(self.url, self.wsdl, cert_path=self.cert, key_path=self.key)

        with patch.object(downloader, '_get_statement_urls', autospec=True) as get_statement_urls:
            get_statement_urls.return_value = ['url_1.test']
            with patch('requests.get') as requests_get:
                requests_get.return_value.ok = True
                requests_get.return_value.text = 'content'
                list(downloader.get_statements(self.start_date, self.end_date))
        requests_get.assert_called_with('url_1.test', cert=(str(self.cert), str(self.key)))

    @patch('zeep.Client', autospec=True)
    def test_create_service(self, mock_client):
        downloader = CSOBStatementDownloader(self.url, self.wsdl, key_path=self.key, cert_path=self.cert)

        mock_client.return_value.create_service.return_value = sentinel.service

        result = downloader._create_service()

        self.assertEqual(sentinel.service, result)
        self.assertEqual(str(self.wsdl), mock_client.call_args[0][0])
        self.assertEqual((str(self.cert), str(self.key)), mock_client.call_args[1]['transport'].session.cert)

        mock_client.return_value.create_service.assert_called_with('{http://ceb-bc.csob.cz/CEBBCWS}CEBBCWSBinding',
                                                                   self.url)

    def test_get_statement_urls(self):
        file_list = [self.FileDetail(Type='VYPIS', Status='D', Filename='file_1.xml', Url='https://bank.test/file_1'),
                     self.FileDetail(Type='VYPIS', Status='D', Filename='file_2.xml', Url='https://bank.test/file_2'),
                     self.FileDetail(Type='VYPIS', Status='D', Filename='file_3.txt', Url='https://bank.test/file_3'),
                     self.FileDetail(Type='VYPIS', Status='R', Filename='file_4.xml', Url='https://bank.test/file_4')]
        service_mock = Mock()
        service_mock.GetDownloadFileList_v2.return_value.FileList.FileDetail = file_list

        downloader = CSOBStatementDownloader(self.url, self.wsdl, contract_number=self.contract)
        with patch.object(downloader, '_create_service', return_value=service_mock, autospec=True):
            with self.assertLogs() as log:
                result = downloader._get_statement_urls(self.start_date, self.end_date)
            self.assertEqual(log.output, ['WARNING:teller.downloaders:Could not get file file_4.xml - file not ready.'])

        filter_criteria = {'FileTypes': ['VYPIS'],
                           'CreatedAfter': self.start_date,
                           'CreatedBefore': self.end_date + timedelta(days=1)}
        service_mock.GetDownloadFileList_v2.assert_called_with(self.contract, Filter=filter_criteria)
        self.assertEqual(['https://bank.test/file_1', 'https://bank.test/file_2'], result)

    def test_get_statement_urls_service_error(self):
        service_mock = Mock()
        service_mock.GetDownloadFileList_v2.side_effect = Exception

        downloader = CSOBStatementDownloader(self.url, self.wsdl, contract_number=self.contract)
        with patch.object(downloader, '_create_service', autospec=True) as create_service:
            create_service.return_value = service_mock
            with self.assertRaisesRegex(TellerDownloadError, 'Could not obtain file list'):
                downloader._get_statement_urls(self.start_date, self.end_date)

    def test_get_statement_urls_file_error(self):
        file_list = [self.FileDetail(Type='VYPIS', Status='F', Filename='file_1.xml', Url='https://bank.test/file_1')]
        service_mock = Mock()
        service_mock.GetDownloadFileList_v2.return_value.FileList.FileDetail = file_list

        downloader = CSOBStatementDownloader(self.url, self.wsdl, contract_number=self.contract)
        with patch.object(downloader, '_create_service', autospec=True) as create_service:
            create_service.return_value = service_mock
            with self.assertRaisesRegex(TellerDownloadError, 'Could not get file file_1.xml - error on the bank side'):
                downloader._get_statement_urls(self.start_date, self.end_date)

    def test_get_statement_urls_unknown_status_error(self):
        file_list = [self.FileDetail(Type='VYPIS', Status='X', Filename='file_1.xml', Url='https://bank.test/file_1')]
        service_mock = Mock()
        service_mock.GetDownloadFileList_v2.return_value.FileList.FileDetail = file_list

        downloader = CSOBStatementDownloader(self.url, self.wsdl, contract_number=self.contract)
        with patch.object(downloader, '_create_service', autospec=True) as create_service:
            create_service.return_value = service_mock
            with self.assertRaisesRegex(TellerDownloadError, 'Unknown status of file file_1.xml'):
                downloader._get_statement_urls(self.start_date, self.end_date)
